import {AdEntity, NewAdEntity, SimpleAdEntity} from "../types";
import {ValidationError} from "../utils/errors";
import {pool} from "../utils/db";
import {FieldPacket} from "mysql2";
import {v4 as uuid} from 'uuid';

type AdRecordResults = [AdEntity[], FieldPacket[]];

export class AdRecord implements AdEntity {
    public id: string;
    public name: string;
    public description: string;
    public price: number;
    public url: string;
    public lat: number;
    public lon: number;

    //w konstruktorze uzywamy jako argumentu dodatkowego interfejsu
    //utworzonego wyżej
    constructor(obj: NewAdEntity) {
        //walidacja danych - również pod kątem pojemności pól w DB
        if(!obj.name || obj.name.length > 100) {
            throw new ValidationError('Nazwa ogłoszenia nie może być pusta, ani przekraczać 100 znaków.');
        }

        if(obj.description.length > 1000) {
            throw new ValidationError('Treść ogłoszenia nie może przekraczać 1000 znaków.');
        }

        if(obj.price < 0 || obj.price > 9999999) {
            throw new ValidationError('Cena nie może być mniejsza niż 0 lub większa niż 9 999 999.');
        }

        //@TODO: Check if URL is valid
        if(!obj.url || obj.url.length > 100) {
            //tymczasowo
            throw new ValidationError('Link ogłoszenia nie może być pusty, ani przekraczać 100 znaków.');
        }

        if(typeof obj.lat !== 'number' || typeof obj.lon !== 'number') {
            throw new ValidationError('Nie można zlokalizować ogłoszenia.');
        }

        this.id = obj.id;
        this.name = obj.name;
        this.description = obj.description;
        this.price = obj.price;
        this.url = obj.url;
        this.lat = obj.lat;
        this.lon = obj.lon;
    }

    static async getOne(id: string): Promise<AdRecord | null> {
        const [results] = await pool.execute("SELECT * FROM `ads` WHERE id = :id", {
            id: id,
        }) as AdRecordResults;

        return results.length === 0 ? null : new AdRecord(results[0]);
    }

    static async findAll(name: string): Promise<SimpleAdEntity[]> {
        const [results] = await pool.execute("SELECT * FROM `ads` WHERE `name` LIKE :search", {
            search: `%${name}%`,
        }) as AdRecordResults;

        //nie mapujemy tu na tablicę nowych obiektów AdRecord z użyciem new
        //ale destrukturyzujemy przefiltrowane składniki i zwracamy je
        //jako obieky w zwracanej tablicy (tablicy przefiltrowanej)
        //interfejs w Promisie do pilnowania poprawności zwracanych danych
        return results.map(item => {
            const {id, lat, lon} = item;
            return {id, lat, lon};
        });
    }

    async insert(): Promise<void> {
        if(!this.id) {
            this.id = uuid();
        } else {
            throw new Error('Cannot insert something that is already inserted');
        }

        const SQL = "INSERT INTO `ads`(`id`, `name`, `description`, `price`, `url`, `lat`, `lon`)" +
            " VALUES(:id, :name, :description, :price, :url, :lat, :lon)";
        await pool.execute(SQL, this);

    }
}